{ 2004-03-13 programmeerimisvistluse edasijudnute 3. lesande lahendus    }
{ lesanne: KALLISKIVID                                                     }
{ Autor: Kristo Tammeoja                                                    }
{ Testitud: FPC 1.0.10 / Win32                                              }

{ O(N^3) }

const
    SIS_FAIL = 'KK.SIS';
    VAL_FAIL = 'KK.VAL';
    MAX_N = 1000;

type
    t_kalliskivi = record
        nr : longint;
        x, y, h : longint;
    end;

var
    N : longint;
    kalliskivi : array[0..MAX_N] of t_kalliskivi;

    // Massiivide suurim_kasum ja parim_eelmine leidmisel arvestatakse
    // vaid neid kalliskive, mis mravad maardla phja kuju,
    // st mne teise kalliskivi kaeveaugus olevat kivi ei arvestata.

    // l-is element nitab suurimat kasumit, mis on vimalik saada,
    // kui l-is kalliskivi on kige lunapoolsem (st kige suurema
    // x-koordinaadiga) kalliskivi, mis maardlast vlja kaevati...
    suurim_kasum : array[0..MAX_N] of double;
    // ...nitab, milline pidi sellisel juhul olema temale eelnev
    // phjapoolne kalliskivi.
    parim_eelmine : array[0..MAX_N] of longint;

    // Mrgitud on kik kivid, mis tuleb vlja kaevata
    vastus : array[1..MAX_N] of boolean;
    // Vastuses olevate kivide arv
    vastus_cnt : longint;

//-------------------- SORTEERIMISEKS -----------------------

// Viime suuremat elementi les
procedure yles (i : longint);
var
    l : longint;
    kk : t_kalliskivi;
begin
    if i = 1 then exit;

    l := i div 2;

    if kalliskivi[i].x > kalliskivi[l].x then begin
        kk := kalliskivi[i];
        kalliskivi[i] := kalliskivi[l];
        kalliskivi[l] := kk;

        yles (l);
    end;
end; //yles

// Viime viksemat elementi alla.
procedure alla (i : longint; kuhja_suurus : longint);
var
    k, l : longint;
    kk : t_kalliskivi;
begin
    l := i;

    k := 2*i;
    if k <= kuhja_suurus then
        if kalliskivi[k].x > kalliskivi[l].x then
            l := k;

    k := 2*i + 1;
    if k <= kuhja_suurus then
        if kalliskivi[k].x > kalliskivi[l].x then
            l := k;

    if i <> l then begin
        kk := kalliskivi[i];
        kalliskivi[i] := kalliskivi[l];
        kalliskivi[l] := kk;

        alla (l, kuhja_suurus);
    end;
end; //alla

// Sorteerib kivid x-i jrgi kasvavas jrjekorras.
procedure sorteeri_kivid;
var
    l : longint; // kuhja pikkus
    kk : t_kalliskivi;
begin
    // Teeme kuhja
    for l := 2 to N do
        yles (l);

    // Vtame kuhja laiali
    for l := N downto 2 do begin
        kk := kalliskivi[1];
        kalliskivi[1] := kalliskivi[l];
        kalliskivi[l] := kk;
        // kuhja ji nd l-1 kivi, korrastame kuja
        alla (1, l-1);
    end; //for l
end; //sorteeri_kivid

//-------------------- LAHENDAMINE ------------------------

procedure loe_andmed;
var
    fail : text;
    l :  longint;
begin
    assign (fail, SIS_FAIL); reset (fail);

    readln (fail, N);

    for l := 1 to N do begin
        kalliskivi[l].nr := l;
        readln (fail, kalliskivi[l].x, kalliskivi[l].y, kalliskivi[l].h);
    end;

    close (fail);
end; //loe_andmed

// Tagastab TRUE, kui kalliskivi a on tielikult kallsikivi
// b kaeveaugus.
function a_on_b_kaeveaugus (a, b : longint) : boolean;
var
    va, vb, pa, pb : longint;
begin
    // leiame nii kalliskivi a kui ka b jaoks nii tema kaeveaugu
    // vasakpoolseima (viksema x-koordinaadiga) ja parempoolseima
    // otspunkti x-koordinaadi
    va := kalliskivi[a].x - kalliskivi[a].y;
    pa := kalliskivi[a].x + kalliskivi[a].y;
    vb := kalliskivi[b].x - kalliskivi[b].y;
    pb := kalliskivi[b].x + kalliskivi[b].y;

    // a on tielikult b kaeveaugus kui tema kaeveaugu otspunktid
    // jvad tielikult b kaeveaugu otspunktide vahele
    a_on_b_kaeveaugus := (vb <= va) and (pa <= pb);
end; //a_on_b_kaeveaugus

// Tagastab tiendava kaevamisele kuuluva summa, kui kalliskivi
// a jrel (paremale e lunapoole) hakatakse kaevama kalliskivi b.
function b_lisahind (a, b : longint) : double;
var
    hind : double;
    pa, vb : longint;
begin
    // leiame ainult kalliskivi b kaeveaugu tegemise hinna
    hind := sqr(kalliskivi[b].y);

    // leiame kalliskivi a kaeveaugu parempoolseima punkti
    // ja kalliskivi b kaeveaugu vasakpoolseima tipu x-koordinaadi
    pa := kalliskivi[a].x + kalliskivi[a].y;
    vb := kalliskivi[b].x - kalliskivi[b].y;

    // kui kaeveaukudel on hisosa, siis lahutame selle maha
    if vb < pa then
        hind := hind - sqr(double(pa-vb)) / 4;

    b_lisahind := hind;
end; //b_lisahind

procedure lahenda;
var
    esimene : boolean;
    kivi, eelmine, lisakivi : longint;
    kivi_kasum : double;
    // parima tulemuse leidmiseks
    parim_kivi : longint;
    // minimaalne kaeveaugu vasak serv
    min_v_serv : longint;
begin
    // sorteerime kalliskivid x-i jrgi kasvavalt
    sorteeri_kivid;

    // lisame kalliskivid massiivi algusesse veel he pseudokivi
    min_v_serv := kalliskivi[1].x - kalliskivi[1].y;
    for kivi := 1 to N do
        if kalliskivi[kivi].x - kalliskivi[kivi].y < min_v_serv then
            min_v_serv := kalliskivi[kivi].x - kalliskivi[kivi].y;
    kalliskivi[0].x := min_v_serv - 1;
    kalliskivi[0].y := 0;
    kalliskivi[0].h := 0;

    // baaslesanne, kui htegi kivi vlja ei kaeva (st kaevab vlja nn
    // pseudokivi), siis on kasum 0
    suurim_kasum[0] := 0;
    parim_eelmine[0] := 0;

    // Hakkame suurimat kasumit arvutama lejnud kalliskivide jaoks
    for kivi := 1 to N do begin
        // Mrgime ra, et me pole veel htegi varianti lbi vaadanud
        esimene := true;
        suurim_kasum[kivi] := 0;

        // Vaatame lbi kik eelnevad kivid
        for eelmine := 0 to kivi-1 do begin
            // Kontrollime, et mlemad oleksid olulised maardla phja
            // kuju jaoks
            if    a_on_b_kaeveaugus (kivi, eelmine)
               or a_on_b_kaeveaugus (eelmine, kivi)
            then
                continue;

            // kui kivid sobisid, siis leiame kasumi
            kivi_kasum := // kasum kuni kivini 'eelmine' kaevamisel
                            suurim_kasum[eelmine]
                          // kaevamisel tekkiv lisakulu
                          - b_lisahind (eelmine, kivi);
            // lisaks veel need kalliskivid, mis jvad 'kivi' kaeveauku,
            // kuid pole kalliskivi 'eelmine' kaeveaugus
            // (kaasa arvatud kalliskivi 'kivi' ise)
            for lisakivi := eelmine+1 to N do
                if     a_on_b_kaeveaugus (lisakivi, kivi)
                   and not a_on_b_kaeveaugus (lisakivi, eelmine)
                then
                    kivi_kasum := kivi_kasum + kalliskivi[lisakivi].h;

            // kui nd saame suurema kasumi kui varem
            if    esimene
               or (kivi_kasum > suurim_kasum[kivi])
            then begin
                suurim_kasum[kivi] := kivi_kasum;
                parim_eelmine[kivi] := eelmine;
                esimene := false;
            end;
        end; //for eelmine
    end; //for kivi

    // Leiame nd, millisel juhul oli tulemus parim
    parim_kivi := 0;
    for kivi := 1 to N do
        if suurim_kasum[kivi] > suurim_kasum[parim_kivi] then
            parim_kivi := kivi;

    // Hakkame vastust kirjutama
    for kivi := 1 to N do
        vastus[kivi] := false;

    // Kime nd lbi need kivid, mis mravad maardla phja kuju
    // Tuleb thele panna, et vastusesse on vaja kivide numbreid
    // enne sorteerimist
    kivi := parim_kivi;
    while kivi <> 0 do begin
        vastus[kalliskivi[kivi].nr] := true;
        // Lisaks veel kivid, mis asuvad 'kivi' kaeveaugus
        for lisakivi := 1 to N do
            if a_on_b_kaeveaugus (lisakivi, kivi) then
                vastus[kalliskivi[lisakivi].nr] := true;
        // Vtame jrgmise kivi vastusest
        kivi := parim_eelmine[kivi];
    end;

    // Loeme vastusesse minevate kivide arvu ka le
    vastus_cnt := 0;
    for kivi := 1 to N do
        if vastus[kivi] then
            vastus_cnt := vastus_cnt + 1;
end; //lahenda

procedure kirjuta_vastus;
var
    fail : text;
    l : longint;
begin
    assign (fail, VAL_FAIL); rewrite (fail);

    writeln (fail, vastus_cnt);
    for l := 1 to N do
        if vastus[l] then
            write (fail, l, ' ');

    close (fail);
end; //kirjuta_vastus

//----------------------- PEAPROGRAMM -----------------------

begin
    loe_andmed;
    lahenda;
    kirjuta_vastus;
end.
