{ 2004-03-13 programmeerimisvistluse edasijudnute 3. lesande tester      }
{ lesanne: KALLISKIVID                                                     }
{ Autor: Kristo Tammeoja                                                    }

const
    MAX_N = 1000;

type
    t_kalliskivi = record
        nr : longint;
        x, y, h : longint;
        p_serv, v_serv : longint;
        sort : longint;
    end;

var
    N : longint;
    kalliskivi : array[0..MAX_N] of t_kalliskivi;

    // Massiivide suurim_kasum ja parim_eelmine leidmisel arvestatakse
    // vaid neid kalliskive, mis mravad maardla phja kuju,
    // st mne teise kalliskivi kaeveaugus olevat kivi ei arvestata.

    // l-is element nitab suurimat kasumit, mis on vimalik saada,
    // kui l-is kalliskivi on kige lunapoolsem (st kige suurema
    // x-koordinaadiga) kalliskivi, mis maardlast vlja kaevati...
    suurim_kasum : array[0..MAX_N] of double;
    // ...nitab, milline pidi sellisel juhul olema temale eelnev
    // phjapoolne kalliskivi.
    parim_eelmine : array[0..MAX_N] of longint;

    arvestatud_kive : array[1..MAX_N] of longint;
    kive_yleval : array[1..MAX_N] of longint;

    index_parem : array[0..MAX_N] of longint;
    index_vasak : array[0..MAX_N] of longint;

    // Mrgitud on kik kivid, mis tuleb vlja kaevata
    vastus : array[1..MAX_N] of boolean;
    // Vastuses olevate kivide arv
    vastus_cnt : longint;

    // Kontrollimiseks
    oige_vastus_kasum : double;
    vastus_kasum : double;

//-------------------- SORTEERIMISEKS -----------------------

// Viime suuremat elementi les
procedure yles (i : longint);
var
    l : longint;
    kk : t_kalliskivi;
begin
    if i = 1 then exit;

    l := i div 2;

    if kalliskivi[i].sort > kalliskivi[l].sort then begin
        kk := kalliskivi[i];
        kalliskivi[i] := kalliskivi[l];
        kalliskivi[l] := kk;

        yles (l);
    end;
end; //yles

// Viime viksemat elementi alla.
procedure alla (i : longint; kuhja_suurus : longint);
var
    k, l : longint;
    kk : t_kalliskivi;
begin
    l := i;

    k := 2*i;
    if k <= kuhja_suurus then
        if kalliskivi[k].sort > kalliskivi[l].sort then
            l := k;

    k := 2*i + 1;
    if k <= kuhja_suurus then
        if kalliskivi[k].sort > kalliskivi[l].sort then
            l := k;

    if i <> l then begin
        kk := kalliskivi[i];
        kalliskivi[i] := kalliskivi[l];
        kalliskivi[l] := kk;

        alla (l, kuhja_suurus);
    end;
end; //alla

// Sorteerib kivid x-i jrgi kasvavas jrjekorras.
procedure sorteeri_kivid;
var
    l : longint; // kuhja pikkus
    kk : t_kalliskivi;
begin
    // Teeme kuhja
    for l := 2 to N do
        yles (l);

    // Vtame kuhja laiali
    for l := N downto 2 do begin
        kk := kalliskivi[1];
        kalliskivi[1] := kalliskivi[l];
        kalliskivi[l] := kk;
        // kuhja ji nd l-1 kivi, korrastame kuja
        alla (1, l-1);
    end; //for l
end; //sorteeri_kivid

//-------------------- LAHENDAMINE ------------------------

procedure loe_andmed;
var
    fail : text;
    l :  longint;
begin
    assign (fail, paramstr(1)); reset (fail);

    readln (fail, N);

    for l := 1 to N do begin
        kalliskivi[l].nr := l;
        readln (fail, kalliskivi[l].x, kalliskivi[l].y, kalliskivi[l].h);
        kalliskivi[l].p_serv := kalliskivi[l].x + kalliskivi[l].y;
        kalliskivi[l].v_serv := kalliskivi[l].x - kalliskivi[l].y;
    end;

    close (fail);
end; //loe_andmed

// Tagastab TRUE, kui kalliskivi a on tielikult kallsikivi
// b kaeveaugus.
function a_on_b_kaeveaugus (a, b : longint) : boolean;
var
    va, vb, pa, pb : longint;
begin
    // leiame nii kalliskivi a kui ka b jaoks nii tema kaeveaugu
    // vasakpoolseima (viksema x-koordinaadiga) ja parempoolseima
    // otspunkti x-koordinaadi
    va := kalliskivi[a].x - kalliskivi[a].y;
    pa := kalliskivi[a].x + kalliskivi[a].y;
    vb := kalliskivi[b].x - kalliskivi[b].y;
    pb := kalliskivi[b].x + kalliskivi[b].y;

    // a on tielikult b kaeveaugus kui tema kaeveaugu otspunktid
    // jvad tielikult b kaeveaugu otspunktide vahele
    a_on_b_kaeveaugus := (vb <= va) and (pa <= pb);
end; //a_on_b_kaeveaugus

// Tagastab tiendava kaevamisele kuuluva summa, kui kalliskivi
// a jrel (paremale e lunapoole) hakatakse kaevama kalliskivi b.
function b_lisahind (a, b : longint) : double;
var
    hind : double;
    pa, vb : longint;
begin
    // leiame ainult kalliskivi b kaeveaugu tegemise hinna
    hind := sqr(kalliskivi[b].y);

    // leiame kalliskivi a kaeveaugu parempoolseima punkti
    // ja kalliskivi b kaeveaugu vasakpoolseima tipu x-koordinaadi
    pa := kalliskivi[a].x + kalliskivi[a].y;
    vb := kalliskivi[b].x - kalliskivi[b].y;

    // kui kaeveaukudel on hisosa, siis lahutame selle maha
    if vb < pa then
        hind := hind - sqr(double(pa-vb)) / 4;

    b_lisahind := hind;
end; //b_lisahind

procedure lahenda;
var
    parem, vasak : longint;
    kive_abi : longint;
    kivi, jargmine, lisakivi : longint;
    kivi_kasum : double;
    // parima tulemuse leidmiseks
    parim_kivi : longint;
    esimene_blokis : boolean;
begin
    // sorteerime kalliskivid v_serv-i jrgi kasvavalt
    for kivi := 1 to N do
        kalliskivi[kivi].sort := kalliskivi[kivi].v_serv;
    sorteeri_kivid;
    for kivi := 1 to N do
        index_vasak[kivi] := kalliskivi[kivi].nr;

    // sorteerime kalliskivid p_serv-i jrgi kasvavalt
    for kivi := 1 to N do
        kalliskivi[kivi].sort := kalliskivi[kivi].p_serv;
    sorteeri_kivid;
    for kivi := 1 to N do
        index_parem[kivi] := kalliskivi[kivi].nr;

    // seame kalliskivid esialgsesse jrjekordavalt
    for kivi := 1 to N do
        kalliskivi[kivi].sort := kalliskivi[kivi].nr;
    sorteeri_kivid;

    // Lisame nn pseudokivi kige vasakpoolsemaks
    kalliskivi[0].nr := 0;
    kalliskivi[0].x := kalliskivi[index_vasak[1]].v_serv - 1;
    kalliskivi[0].y := 0;
    kalliskivi[0].h := 0;
    kalliskivi[0].v_serv := kalliskivi[0].x;
    kalliskivi[0].p_serv := kalliskivi[0].x;

    // Tidame massiivi kive_yleval ja algvrtustame massiivid
    // suurim_kasum ja parim_eelmine;
    for kivi := 1 to N do begin
        kive_yleval[kivi] := 0;
        for lisakivi := 1 to N do
            if a_on_b_kaeveaugus(lisakivi, kivi) then
                kive_yleval[kivi] :=   kive_yleval[kivi]
                                     + kalliskivi[lisakivi].h;
        suurim_kasum[kivi] :=   kive_yleval[kivi]
                              - sqr(kalliskivi[kivi].y);
        parim_eelmine[kivi] := 0;
    end;

    // Algvrtustame massiivi arvestatud_kive
    for vasak := 1 to N do
        arvestatud_kive[vasak] := 0;

    // baaslesanne, kui htegi kivi vlja ei kaeva (st kaevab vlja nn
    // pseudokivi), siis on kasum 0
    suurim_kasum[0] := 0;
    parim_eelmine[0] := 0;

    // lahendamine
    for parem := 1 to N do begin
        // Tidame massiivi 'arvestatud_kive' antud kaeveaugu
        // parema serva jaoks
        if    kalliskivi[index_parem[parem-1]].p_serv
           <> kalliskivi[index_parem[parem]].p_serv
        then begin
            kive_abi := 0;
            for vasak := N downto 1 do begin
                // vaatame
                if vasak = N then esimene_blokis := true
                else esimene_blokis :=
                           kalliskivi[index_vasak[vasak+1]].v_serv
                        <> kalliskivi[index_vasak[vasak]].v_serv;

                // Liidame kokku kik kivi
                if esimene_blokis
                then begin
                    kivi := vasak;
                    while   kalliskivi[index_vasak[vasak]].v_serv
                          = kalliskivi[index_vasak[kivi]].v_serv
                    do begin
                        if   kalliskivi[index_parem[parem]].p_serv
                           = kalliskivi[index_vasak[kivi]].p_serv
                        then
                            kive_abi :=   kive_abi
                                        + kalliskivi[index_vasak[kivi]].h;
                        kivi := kivi - 1;
                    end;
                end;

                arvestatud_kive[vasak] :=   arvestatud_kive[vasak]
                                          + kive_abi;
            end;
        end;

        //
        kivi := index_parem[parem];
        for vasak := 1 to N do begin
            jargmine := index_vasak[vasak];

            if     (kalliskivi[kivi].x < kalliskivi[jargmine].x)
               and not a_on_b_kaeveaugus(kivi, jargmine)
               and not a_on_b_kaeveaugus(jargmine, kivi)
            then begin
                kivi_kasum :=   suurim_kasum[kivi]
                              + kive_yleval[jargmine]
                              - arvestatud_kive[vasak]
                              - b_lisahind(kivi, jargmine);

                if (kivi_kasum > suurim_kasum[jargmine])
                then begin
                    suurim_kasum[jargmine] := kivi_kasum;
                    parim_eelmine[jargmine] := kivi;
                end;
            end;
        end;
    end;

    // Leiame nd, millisel juhul oli tulemus parim
    parim_kivi := 0;
    for kivi := 1 to N do
        if suurim_kasum[kivi] > suurim_kasum[parim_kivi] then
            parim_kivi := kivi;

    // Vastuse kontroll
    oige_vastus_kasum := suurim_kasum[parim_kivi];

end; //lahenda

procedure loe_vastus;
var
    fail : text;
    l : longint;
    K : longint;
    kivi, eelmine : longint;
    jrk : boolean;
begin
    writeln ('vljundafaili lugemine...');

    assign (fail, paramstr(2)); reset (fail);

    readln (fail, K);

    for l := 1 to N do
        vastus[l] := false;

    eelmine := 0;
    jrk := true;
    for l := 1 to K do begin
        read (fail, kivi);

        if (kivi < 1) or (N < kivi) then begin
            writeln ('kivi number pole vahemikus [1..N]');
            halt;
        end;

        if vastus[kivi] then begin
            writeln ('kivi number #', kivi, ' on sisendis topelt');
            halt;
        end;
        vastus[kivi] := true;

        if jrk and (eelmine > kivi) then begin
            writeln ('kivide numbrid pole kasvavas jrjekorras');
            jrk := false;
        end;
        eelmine := kivi;
    end;

    close (fail);
    writeln ('...loetud');

    // kontrollime kasumit
    vastus_kasum := 0;

    // kalliskivide hind
    for l := 1 to N do
        kalliskivi[l].sort := kalliskivi[l].x;
    sorteeri_kivid;

    for kivi := 1 to N do
        if vastus[kalliskivi[kivi].nr] then
            vastus_kasum := vastus_kasum + kalliskivi[kivi].h;

    // kaevamise hind
    eelmine := 0;
    for kivi := 1 to N do
        if vastus[kalliskivi[kivi].nr] then
            for l := 1 to N do begin
                if kivi = l then continue;

                if a_on_b_kaeveaugus(l, kivi) then
                    vastus[kalliskivi[l].nr] := false;
            end;

    for kivi := 1 to N do
        if vastus[kalliskivi[kivi].nr] then begin
            vastus_kasum :=   vastus_kasum
                            - b_lisahind(eelmine, kivi);
            eelmine := kivi;
        end;

    // Vastuse kontroll
    if abs (oige_vastus_kasum - vastus_kasum) < 0.0000001 then begin
        writeln ('OK');
    end else begin
        writeln (oige_vastus_kasum:0:2);
        writeln (vastus_kasum:0:2);
    end;
end; //loe_vastus

//----------------------- PEAPROGRAMM -----------------------

begin
    loe_andmed;
    lahenda;
    loe_vastus;
end.
